//---------------------------------------------------------------
// longrun.c
//
// Longrun driver for Transmeta Crusoe.
//
// Based on FreeBSD driver by: Tamaotsu Hattori
// BeOS Driver by: Anders Granlund.
//---------------------------------------------------------------

/*
 * Copyright (c) 2001 Tamotsu Hattori.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "driver.h"
#include "longrun.h"

//---------------------------------------------------------------
// Globals
//---------------------------------------------------------------
uint8 longrun_enabled = 0;

//---------------------------------------------------------------
// Costants
//---------------------------------------------------------------
#define MSR_TMx86_LONGRUN_STATUS	0x80860007
#define MSR_TMx86_LONGRUN			0x80868010
#define MSR_TMx86_LONGRUN_FLAGS		0x80868011

//---------------------------------------------------------------
// Macros
//---------------------------------------------------------------
#define LONGRUN_MODE_MASK(x) 		((x) & 0x000000007f)
#define LONGRUN_MODE_RESERVED(x) 	((x) & 0xffffff80)
#define LONGRUN_MODE_WRITE(x, y) 	(LONGRUN_MODE_RESERVED(x) | LONGRUN_MODE_MASK(y))

//---------------------------------------------------------------
// MSR information
//---------------------------------------------------------------
union msrinfo {
	uint64	msr;
	uint32	regs[2];
};

//---------------------------------------------------------------
// MSR information for every longrun modes
//---------------------------------------------------------------
uint32 longrun_modes[LONGRUN_MODE_MAX][3] = {
	/*  MSR low, MSR high, flags bit0 */
	{	  0,	  0,		0},		/* LONGRUN_MODE_MINFREQUENCY */
	{	  0,	100,		0},		/* LONGRUN_MODE_ECONOMY */
	{	  0,	100,		1},		/* LONGRUN_MODE_PERFORMANCE */
	{	100,	100,		1},		/* LONGRUN_MODE_MAXFREQUENCY */
};


//---------------------------------------------------------------
// Read MSR (Machine specific registers)
//---------------------------------------------------------------
static inline uint64 rdmsr( u_int msr)
{
	uint64 rv;
	asm volatile(".byte 0x0f, 0x032" : "=A" (rv) : "c" (msr));
	return (rv);
}

//---------------------------------------------------------------
// Write MSR (Machine specific registers)
//---------------------------------------------------------------
static inline void wrmsr( uint msr, uint64 newval)
{
	asm volatile(".byte 0x0f, 0x30" : : "A" (newval), "c" (msr));
}

//---------------------------------------------------------------
// Read cpuid data
//---------------------------------------------------------------
static void cpuid(uint ax, uint * p)
{
	asm volatile(
	".byte	0x0f, 0xa2;"
	"movl	%%eax, (%2);"
	"movl	%%ebx, 4(%2);"
	"movl	%%ecx, 8(%2);"
	"movl	%%edx, 12(%2);"
	:"=a"(ax)
	:"0"(ax), "S"(p)
	:"bx", "cx", "dx"
	);
}

//---------------------------------------------------------------
// longrun_init
// Initialize longrun
//---------------------------------------------------------------
int longrun_init()
{
	// Longrun is enabled
	longrun_enabled = 1;

	// return longrun status
	return longrun_enabled;
}


//---------------------------------------------------------------
// longrun_setmode
// Set longrun mode
//---------------------------------------------------------------
int longrun_setmode( uint8 mode)
{
	union msrinfo	msrinfo;
	
	// Make sure mode is valid
	if( mode >= LONGRUN_MODE_MAX)
		return 0;
	
	// Write longrun mode to msr
	msrinfo.msr     = rdmsr(MSR_TMx86_LONGRUN);
	msrinfo.regs[0] = LONGRUN_MODE_WRITE( msrinfo.regs[0], longrun_modes[mode][0]);
	msrinfo.regs[1] = LONGRUN_MODE_WRITE( msrinfo.regs[1], longrun_modes[mode][1]);
	wrmsr( MSR_TMx86_LONGRUN, msrinfo.msr);
	
	// Write longrun flag to msr
	msrinfo.msr     = rdmsr(MSR_TMx86_LONGRUN_FLAGS);
	msrinfo.regs[0] = ( msrinfo.regs[0] & ~1) | longrun_modes[mode][2];
	wrmsr(MSR_TMx86_LONGRUN_FLAGS, msrinfo.msr);
	
	return 1;
}

//---------------------------------------------------------------
// longrun_getmode
// Get longrun mode
//---------------------------------------------------------------
uint8 longrun_getmode()
{
	union msrinfo	msrinfo;
	uint			low,high,flags;
	uint8			test,mode;	

	// Get longrun registers	
	msrinfo.msr = rdmsr( MSR_TMx86_LONGRUN);
	low  = LONGRUN_MODE_MASK( msrinfo.regs[0]);
	high = LONGRUN_MODE_MASK( msrinfo.regs[1]);
	flags = rdmsr( MSR_TMx86_LONGRUN_FLAGS) & 0x01;

	// We have the longrun registers so try to find
	// out which mode we are currently in.
	mode = LONGRUN_MODE_MAX;
	for( test = 0; test < LONGRUN_MODE_MAX; test++)
	{
		if( low  == longrun_modes[test][0] && 
			high == longrun_modes[test][1] &&
			flags == longrun_modes[test][2])
			{
				mode = test;
				continue;
			}
	}
	
	// Return the current longrun mode	
	return mode;
}


//---------------------------------------------------------------
// longrun_getstatus
// Get longrun CPU status
//---------------------------------------------------------------
int longrun_getstatus( uint *frequency, uint *voltage, uint *percentage)
{
	uint regs[4];

	cpuid( MSR_TMx86_LONGRUN_STATUS, regs);
	*frequency  = regs[0];
	*voltage    = regs[1];
	*percentage = regs[2];
	return 1;
}
